"""
Package metadata
"""
__version_info__ = (1, 0, 31)
__version__ = '.'.join(map(str, __version_info__))
__author__ = "Hyperledger"
__license__ = "Apache 2.0"

__all__ = ['__version_info__', '__version__', '__author__', '__license__']
