LARGE_PUBLIC_RHO = 256
LARGE_PUBLIC_B = 1376
LARGE_PRIME = 1024
LARGE_VPRIME_PRIME = 2724
LARGE_VPRIME = 2128
LARGE_MVECT = 592
LARGE_UTILDE = LARGE_MVECT
LARGE_RTILDE = 672
LARGE_E_START = 596
LARGE_E_END_RANGE = 119
LARGE_NONCE = 80
LARGE_ETILDE = 456
LARGE_VTILDE = 3060
LARGE_MASTER_SECRET = 256
LARGE_ALPHATILDE = 2787
LARGE_M2_TILDE = 1024
ITERATIONS = 4

PAIRING_GROUP = 'SS1024'  # super singular curve, 1024 bits

MASTER_SEC_RAND = "master_secret_rand"
REVEALED_ATTRS = "revealedAttrs"
ZERO_INDEX = "0"

ATTRS = "attrs"
DELTA = "DELTA"
TVAL = "Tval"
KEYS = "keys"
NONCE = "nonce"

APRIME = "Aprime"
EPRIME = "eprime"
VPRIME = "vprime"

EVECT = "evect"
MVECT = "mvect"
VVECT = "vvect"

ETILDE = "etilde"
MTILDE = "mtilde"
VTILDE = "vtilde"

CRED_A = "A"
CRED_E = "e"
CRED_V = "v"

C_VALUE = "c"

PK_R = "R"
PK_N = "N"
PK_S = "S"
PK_Z = "Z"

ISSUER = "issuer"
PROOF = "proof"

NAME = "name"
VERSION = "version"
TYPE = "type"
ATTR_NAMES = "attr_names"

TYPE_CL = "CL"
